;; Generated package description from jinx.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "jinx" "2.10.0.20260813.1" "Enchanted Spell Checker" '((emacs "29.1") (compat "31")) :commit "23f6d028decb6081ee30659e7e841f6046e5f511" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("convenience" "text") :url "https://github.com/minad/jinx")
